\Scroller.xpl
\Based on Polaris's example program.

string 0;			\use the zero-terminated string convention
include	c:\cxpl\codes;		\run-time library routines (intrinsics)

def	ScrWidth = 320,		\screen dimensions (pixels)
	ScrHeight = 200;
def	FontWidth = 32,		\font character dimensions (pixels)
	FontHeight = 64;
def	FontImageWidth = 1536,
	FontImageHeight = 64;

int	CpuReg,			\array containing copy of CPU hardware registers
	DataSeg,		\segment address where our data got loaded
	FrameCount;

char	Message,		\message string to be displayed
	Background(ScrWidth*ScrHeight),	\background screen buffer
	FontImage(FontImageWidth*FontImageHeight),	\image from the .bmp file
	ScrBuffer(ScrWidth*ScrHeight);	\screen buffer



func	StrLen(Str);		\Returns the number of characters in a string
char	Str;
int	I;
begin
I:= 0;
while Str(I) \#0\ do I:= I+1;
return I;
end;	\StrLen



proc	WaitVB;			\Wait for vertical blank
begin
while port($3DA) & $08 do;	\wait for vertical blank to go away
repeat until port($3DA) & $08;	\wait for vertical blank
end;	\WaitVB



proc	LoadBmp(Name, Image);
\Load .bmp file into Image and set up color registers
char	Name,		\file name
	Image;		\array to store image into
int	Hand, I, T,
	X, Y,		\coordinates
	W, H,		\width and height
	R, G, B;	\red, green, blue
begin
Hand:= FOpen(Name, 0);			\open file for input
FSet(Hand, ^I);				\set device 3 to handle
OpenI(3);
for I:= 0, 17 do T:= ChIn(3);		\skip unused header info
W:= ChIn(3) + ChIn(3)<<8;		\18 width
T:= ChIn(3) + ChIn(3)<<8;		\20 unused
H:= ChIn(3) + ChIn(3)<<8;		\22 height
for I:= 24, 53 do T:= ChIn(3);		\skip unused header info

port($3C8):= 0;				\set up color registers
for I:= 0, 255 do
	begin
	B:= ChIn(3)>>2;			\order is backwards
	G:= ChIn(3)>>2;
	R:= ChIn(3)>>2;
	T:= ChIn(3);
	port($3C9):= R;
	port($3C9):= G;
	port($3C9):= B;
	end;

for Y:= -(H-1), 0 do			\.bmp files are upside down!
    for X:= 0, W-1 do
	Image(X-Y*W):= ChIn(3);
FClose(Hand);
end;	\LoadBmp

\-------------------------------------------------------------------------------

proc	MaskedBlit(ImSrc, ImDst, Xs, Ys, Xd, Yd, W, H);
\Copy rectangular area of image
char	ImSrc, ImDst;	\image array source and destination
int	Xs, Ys, Xd, Yd,	\coordinates in source and destination
	W, H;		\width and height of area to copy
int	X, Y, C, Is, Id;
begin
for Y:= 0, H-1 do
	begin
	Is:= (Y+Ys)*FontImageWidth + Xs;
	Id:= (Y+Yd)*ScrWidth + Xd;
	for X:= 0, W-1 do
		begin
		C:= ImSrc(X+Is);
		if C then
		    if X+Xd>=0 & X+Xd<ScrWidth then ImDst(X+Id):= C;
		end;
	end;
end;	\MaskedBlit



proc	DrawChar(X, Y, Ch);
int	X, Y, Ch;
int	I;
char	ChMap;
begin
\Figure out the character mapping for the character. String contains the order
\ of the characters in BIGFONT.BMP. Credit and greets go to Barzoule for this
\ mapping approach.
ChMap:= "abcdefghijklmnopqrstuvwxyz0123456789^"'(),.;:+-!?";

\Character Ch is passed by value, not by reference, so we can force uppercase
\ letters to lower case to match the map.
if Ch>=^A & Ch<=^Z then Ch:= Ch + ^a - ^A;

\Do a linear search
loop	begin
	for I:= 0, StrLen(ChMap)-1 do
		if ChMap(I) = Ch then quit; \(I = index into ChMap)
	quit;				    \in case Ch isn't found (I = StrLen)
	end;

\Copy character from FontImage to ScrBuffer (if we found a valid one)
if I < StrLen(ChMap) then
    if X+FontWidth>0 & X<ScrWidth then	\verify that the X position is visible
	MaskedBlit(FontImage, ScrBuffer,
		I*FontWidth, 0,		\coordinates of letter in FontImage
		X, Y,			\coordinates to put letter in ScrBuffer
		FontWidth, FontHeight); \font character dimensions
end;	\DrawChar



proc	DrawString(X, Y, Msg);
int	X, Y;
char	Msg;
int	I;
begin
for I:= 0, StrLen(Msg)-1 do
	DrawChar(X+I*FontWidth, Y, Msg(I)&$7F);
end;	\DrawString

\-------------------------------------------------------------------------------

begin	\Main
CpuReg:= GetReg;
DataSeg:= CpuReg(12);

SetVid($13);					\320x200x8-bit color
LoadBmp("BIGFONT.BMP", FontImage);
LoadBmp("BACKGND.BMP", Background);		\use palette from Backgnd

Message:= "Greetz to all you XPL0 coders! ... Have a great time!! - Boreal";
FrameCount:= 0;
repeat	Blit(DataSeg, Background, DataSeg, ScrBuffer, ScrWidth*ScrHeight);
	DrawString(ScrWidth-FrameCount, 68, Message);
	if StrLen(Message)*FontWidth - FrameCount + ScrWidth <= 0 then
		FrameCount:= 0;
	FrameCount:= FrameCount+1;
	WaitVB;
	Blit(DataSeg, ScrBuffer, $A000, 0, ScrWidth*ScrHeight); 
until	ChkKey;

SetVid($03);					\restore normal text mode
end;	\Main
