\Torus2.xpl	3-Feb-2007	Boreal
\3D Wireframe Torus Rotatable with Mouse

include	c:\cxpl\codesi;		\intrinsic code definitions

def	VMode=$12,		\video mode: 640x480x4 graphiss
	TVW=640, TVH=480;	\screen width and height, in pixels

int	CpuReg;			\address of CPU register array (from GetReg)

def	Bel=$07, CR=$0D, Esc=$1B;				\control chars
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\scan codes
def	Pi = 3.14159265358979323846;

\-------------------------------------------------------------------------------

func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
begin
CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
SoftInt(Int);
return CpuReg(0);		\return value in AX register
end;	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl+C, and no cursor
SC:= CallInt($16, $0000);	\BIOS function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
end;	\GetKey



func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer is set to center of screen but is hidden
CallInt($21, $3533);		\Make sure mouse vector ($33) points to something
if CpuReg(1)=0 & CpuReg(11)=0 then return false;
return CallInt($33, $0000);	\reset mouse and return status
end;	\OpenMouse

\-------------------------------------------------------------------------------

proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
def	X, Y, Z;		\dimensions
begin
V3(X):= V1(Y)*V2(Z) - V1(Z)*V2(Y);
V3(Y):= V1(Z)*V2(X) - V1(X)*V2(Z);
V3(Z):= V1(X)*V2(Y) - V1(Y)*V2(X);
end;	\CrossProd



proc	Rotate(V, R, P, W); \3D rotate vector V
real	V,		\3D vector is 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	X, Y, Z, SW, SP, SR, CW, CP, CR, T;
begin
X:= V(0);   Y:= V(1);	Z:= V(2);	\get vector components
SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= Y*CR - Z*SR;		\		       	 
Z:= Z*CR + Y*SR;		\		      1   0   0 
Y:= T;				\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= X*CP + Z*SP;		\		       	 
Z:= Z*CP - X*SP;		\		      CP  0  -SP
X:= T;				\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= X*CW - Y*SW;		\		       	  
Y:= Y*CW + X*SW;		\		       CW  SW	0 
X:= T;				\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
				\		       	  
V(0):= X;   V(1):= Y;	V(2):= Z;	\return values
end;	\Rotate

\===============================================================================

proc	MoveTorus;	\Display and move a torus (a donut)
def	Max=32,		\maximum limits +1 for Sectors and Sections
	Scale = 1.4,	\magnification factor (adjusted to fill screen)
	TorR1 = 100.0,	\distance from the center of the hole in the middle
			\ to the center of the eatable tube-shaped part
	TorR2 = 60.0;	\distance from the center of the eatable tube-shaped
			\ part to the frosting on top
int	Sectors,	\number of slices of the torus (tubes)
	Sections;	\number of slices of a sector (patches)
real	TorX(Max,Max), TorY(Max,Max), TorZ(Max,Max), \coordinates of vertexes
	Ref1(3), Ref2(3), Ref3(3), \reference frame (unit vectors)
	Ref1Z, Ref2Z;	\used to detect movement, which redraws torus
int	RTorX(Max,Max), RTorY(Max,Max);	\rotated coordinates of vertexes

\-------------------------------------------------------------------------------

proc	DrawTorus;	\Draw the torus
\Inputs: Sectors, Sections, RTorX, RTorY.
int	I, J;		\indexes
def	CX=TVW/2+80, CY=TVH/2;
def	Color=3;	\cyan
begin
for J:= 0, Sections-1 do
	begin
	Move(RTorX(0,J)+CX, RTorY(0,J)+CY);
	for I:= 1, Sectors do		\(end point duplicates first point)
		Line(RTorX(I,J)+CX, RTorY(I,J)+CY, Color);
	end;

for I:= 0, Sectors-1 do
	begin
	Move(RTorX(I,0)+CX, RTorY(I,0)+CY);
	for J:= 1, Sections do		\(end point duplicates first point)
		Line(RTorX(I,J)+CX, RTorY(I,J)+CY, Color);
	end;
end;	\DrawTorus

\-------------------------------------------------------------------------------

proc	RotateTorus;	\Rotate torus by moving the mouse
\Inputs: Sectors, Sections, TorX, TorY, TorZ.
\Outputs: RTorX, RTorY.
int	I, J;			\indexes
real	AngX, AngY, AngZ,	\rotation angle about indicated axis (radians)
	Rxx, Rxy, Rxz,		\rotation matrix
	Ryx, Ryy, Ryz,
	Rzx, Rzy, Rzz;
def	X, Y, Z;		\dimensions
begin
CallInt($33, $0B);		\read mouse motion counters
\Mouse X motion rotates object about Y axis
AngY:= Float(-CpuReg(2)) / 500.0;
\Mouse Y motion rotates object about X axis
AngX:= Float(CpuReg(3)) / 500.0;
AngZ:= 0.0;

\STEP 1:
\Rotate reference frame according to changes in mouse position.
\Independently rotating P1 and P2 can cause them to drift relative to
\ each other, but the amount of drift is insignificant because reals are
\ accurately calculated to 16 decimal places.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\STEP 2:
\Rotate each vertex in the object so that it aligns with the reference frame:
\Replace subscripted variables for efficiency and neatness:
Rxx:= Ref1(X);	 Rxy:= Ref1(Y);   Rxz:= Ref1(Z);
Ryx:= Ref2(X);	 Ryy:= Ref2(Y);   Ryz:= Ref2(Z);
Rzx:= Ref3(X);	 Rzy:= Ref3(Y);   Rzz:= Ref3(Z);

\This multiplies each point at its initial position (Po) by the rotation
\ matrix to get the rotated position
\					     		     
\		     	 		     Rxx  Rxy  Rxz 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ryx  Ryy  Ryz 
\		     	 		     Rzx  Rzy  Rzz 
\					     		     
\
for I:= 0, Sectors do			\for all of the sectors+1...
    for J:= 0, Sections do		\for all of the sections+1...
	begin				\(end points duplicate first points)
	RTorX(I,J):= Fix((TorX(I,J)*Rxx + TorY(I,J)*Ryx + TorZ(I,J)*Rzx)*Scale);
	RTorY(I,J):= Fix((TorX(I,J)*Rxy + TorY(I,J)*Ryy + TorZ(I,J)*Rzy)*Scale);
	end;
end;	\RotateTorus

\-------------------------------------------------------------------------------

proc	CalcTorus;
\Find the coordinates of the vertexes of facets that approximate a torus.
\Chop the surface of the torus up into facets (polygons). To simplify later
\ calculations, the last coordinates duplicate the first coordinates. The
\ torus is initially in the X,Y plane (like a donut floating in hot grease).
\Inputs: Sectors, Sections, TorR1, TorR2.
\Outputs: TorX, TorY, TorZ.
int	I, J;
real	A,		\angular increments about the hole in the center
	B,		\angular increments about the center of the tube
	SA, CA, SB, CB;	\sines and cosines of A and B
begin
for I:= 0, Sectors do			\for all of the sectors+1...
	begin				\(end point duplicates first point)
	A:= Float(I) / Float(Sectors) * Pi * 2.0;
	SA:= Sin(A);   CA:= Cos(A);
	for J:= 0, Sections do		\for all of the sections+1...
		begin			\(end point duplicates first point)
		B:= Float(J) / Float(Sections) * Pi * 2.0;
		SB:= Sin(B);   CB:= Cos(B);
		TorX(I,J):= (TorR1 + TorR2*CB) * CA;
		TorY(I,J):= (TorR1 + TorR2*CB) * SA;
		TorZ(I,J):= TorR2 * SB;
		end;
	end;
end;	\CalcTorus

\-------------------------------------------------------------------------------

proc	Reset;	\Initialize variables to starting configuration
def	X, Y, Z;		\dimensions
begin
Ref1(X):= 1.0;   Ref1(Y):= 0.0;   Ref1(Z):= 0.0;
Ref2(X):= 0.0;   Ref2(Y):= 1.0;   Ref2(Z):= 0.0;
Sectors:= 13;    Sections:= 13;
CalcTorus;
end;	\Reset

\-------------------------------------------------------------------------------

begin	\MoveTorus
Reset;

Ref1Z:= 2.0;	\set to an impossible value so that the screen gets redrawn
loop	begin
	RotateTorus;

	if Ref1(2)#Ref1Z ! Ref2(2)#Ref2Z then
		begin
		Ref1Z:= Ref1(2);   Ref2Z:= Ref2(2);
		Clear;
		Text(6, "Move mouse
to rotate


Sectors
[^[] [ ]


Segments
[^X] [^Y]


Reset
[Enter]


Exit
[Esc]");

\Insert arrow character (^Z looks like an EOF, which terminates the compile)
		Cursor(5,5);   ChOut(6, $1A);

	 	DrawTorus;
		end;

	if ChkKey then
		begin
		Ref1Z:= 2.0;	\set so that screen gets redrawn
		case GetKey of
		  -LtArrow:
			if Sectors > 3 then
				[Sectors:= Sectors-1; CalcTorus];
		  -RtArrow:
			if Sectors < Max-1 then	\(end point duplicates 1st one)
				[Sectors:= Sectors+1; CalcTorus];
		  -UpArrow:
			if Sections < Max-1 then
				[Sections:= Sections+1; CalcTorus];
		  -DnArrow:
			if Sections > 3 then
				[Sections:= Sections-1; CalcTorus];
		  CR:	Reset;

		  Esc:	quit
		other	ChOut(0, Bel);
		end;
	end;
end;	\MoveTorus

\===============================================================================

begin	\Main
CpuReg:= GetReg;		\get address of array to access CPU registers
if not OpenMouse then [Text(0, "A mouse is required");   exit];
SetVid(VMode);
MoveTorus;
SetVid(3);			\restore normal text mode
end;	\Main
