\Torus4.xpl	25-Jan-2007	Boreal
\Rotatable 3D Flat-Shaded Torus
\Compile with 32-bit XPL (XPLPX, xpx.bat)
\
\A right-hand coordinate system is used:
\
\				|   Z  (into screen)
\				|  /
\				| /
\				|/
\			--------+-------- X
\			       /|
\			      / |
\			     /	|
\				|
\				Y

include	c:\cxpl\codes;		\intrinsic code definitions (including Paint)

def	VMode=$101,		\video mode: 640x480x8 graphics
	TVW=640, TVH=480,	\screen width and height, in pixels
	ImW=TVH, ImH=TVH;	\image width and height, in pixels (a square)
def	IntSize=4;		\number of bytes in an integer (2 or 4)
def	X, Y, Z;		\dimensions (indexes)

int	CpuReg;			\address of CPU register array (from GetReg)
char	Image(ImW*ImH);		\image buffer, gets copied to the screen

def	Bel=$07, CR=$0D, Esc=$1B;				\control chars
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\scan codes
def	Pi = 3.14159265358979323846;

\-------------------------------------------------------------------------------

func	WExtend(N);		\Sign extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
begin
CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
SoftInt(Int);
return WExtend(CpuReg(0));	\return value in AX register (as 32 bits)
end;	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl+C, and no cursor
SC:= CallInt($16, $0000);	\BIOS function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
end;	\GetKey



func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer is set to center of screen but hidden
CallInt($21, $3533);		\Make sure mouse vector $33 points to something
if WExtend(CpuReg(1))=0 & WExtend(CpuReg(11))=0 then return false;
CallInt($33, $21);		\reset mouse for Windows XP
CallInt($33, $0B);		\read motion counters to reset them under Win XP
return CallInt($33, $0000);	\h/w reset mouse and return status
end;	\OpenMouse



func	MouseButton;		\Return 'true' if a mouse button is down
begin
CallInt($33, $0003);
return CpuReg(1) & $03;
end;	\GetMouseButton

\-------------------------------------------------------------------------------

proc	VectSub(V1, V2, V3);	\Subtract two 3D vectors
real	V1, V2, V3;		\V3:= V1 - V2
begin
V3(X):= V1(X) - V2(X);
V3(Y):= V1(Y) - V2(Y);
V3(Z):= V1(Z) - V2(Z);
end;	\VectSub



proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
begin
V3(X):= V1(Y)*V2(Z) - V1(Z)*V2(Y);
V3(Y):= V1(Z)*V2(X) - V1(X)*V2(Z);
V3(Z):= V1(X)*V2(Y) - V1(Y)*V2(X);
end;	\CrossProd



func real DotProd(V1, V2);	\Return the dot product of two 3D vectors
real	V1, V2;			\V1  V2
return V1(X)*V2(X) + V1(Y)*V2(Y) + V1(Z)*V2(Z);



proc	Unitize(V);		\Convert a 3D vector to its unit vector
real	V;			\vector to convert
real	M;			\magnitude of vector
begin
M:= Sqrt( V(X)*V(X) + V(Y)*V(Y) + V(Z)*V(Z) );
V(X):= V(X)/M;
V(Y):= V(Y)/M;
V(Z):= V(Z)/M;
end;	\Unitize



proc	Rotate(V, R, P, W); \3D rotate vector V
real	V,		\3D vector is 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	SW, SP, SR, CW, CP, CR, T;
begin
SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= V(Y)*CR - V(Z)*SR;		\		       	 
V(Z):= V(Z)*CR + V(Y)*SR;	\		      1   0   0 
V(Y):= T;			\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= V(X)*CP + V(Z)*SP;		\		       	 
V(Z):= V(Z)*CP - V(X)*SP;	\		      CP  0  -SP
V(X):= T;			\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= V(X)*CW - V(Y)*SW;		\		       	  
V(Y):= V(Y)*CW + V(X)*SW;	\		       CW  SW	0 
V(X):= T;			\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
end;	\Rotate					       	  



proc	EraseImage;		\Erase the Image array
int	I;
for I:= 0, ImW*ImH-1 do Image(I):= 0;

\-------------------------------------------------------------------------------

proc	DrawPoly(X0, Y0, VT, Color);	\Draw a filled polygon onto Image
\This works for convex as well as concave polygons, and even for polygons
\ containing holes. Initially a table is set up that defines the coordinates
\ along the edges of the polygon. Each location in the table corresponds to a
\ Y coordinate, and (in the simple, usual case) it contains the corresponding
\ left and right X coordinates of the polygon's edges. After the table is set
\ up, horizontal lines are drawn to fill the polygon. Ref: "Computer Graphics:
\ Principals and Practice, Second Edition" by John Foley, et al.

int	X0, Y0,		\position to draw polygon (pixels)
	VT,		\vertex table: 1-dimensional array of X,Y coordinates
			\ first value gives the number of vertexes (X,Y pairs)
	Color;		\fill color
int	N, HeadTbl, Ptr, MemPtr, \used to build linked-list table
	Gx, Gy, Hx, Hy,	\coordinates of end points of drawn horizontal scan line
	I, J,		\indexes
	X, Y,		\coordinates used to draw horizontal scan lines
	Min, Max,	\minimum and maximum Y coordinates of polygon
	Height;		\height of polygon, in pixels
def	Nil= $80000000;	\linked list null pointer


	proc	BuildLine(I0, I1);
	\Build a table at MemPtr that contains X coordinates along the polygon's
	\ edge. This makes a linked list for each horizontal line that will be
	\ drawn to fill the polygon.
	int	I0, I1;	\indexes in VT specifying vertexes to build line between
	int	X, Y, DX, DY, M, T, Q;
	def	SS = 16; \scale shift factor
	begin
	Gx:= VT(I0) + X0;   Gy:= VT(I0+1) + Y0;	\get coordinates of end points
	Hx:= VT(I1) + X0;   Hy:= VT(I1+1) + Y0;	\ of the polygon's edge

	if Gy > Hy then
		begin				\swap end points, so that Hy is
		T:= Gy;   Gy:= Hy;   Hy:= T;	\ lower than (or equal to) Gy
		T:= Gx;   Gx:= Hx;   Hx:= T;
		end;

	DY:= Hy - Gy;				\delta Y
	if DY = 0 then return;			\horizontal lines can be ignored
	DX:= Hx - Gx;				\ (it's good thing too!)
	M:= DX<<SS / DY;			\scaled slope (rotated 90 deg)

	X:= Gx<<SS;				\scaled starting X coordinate
	for Y:= Gy-Min, Hy-Min-1 do		\follow the edge downward...
		begin				\(bottom-most point isn't drawn)
		T:= X>>SS;			\unscale X coordinate
		Ptr:= HeadTbl + Y*IntSize;	\pointer from list header table
		loop	begin
			if Ptr(0)=Nil ! Ptr(0,1)>T then
				begin		\insert item at front of list
				Q:= MemPtr;	\ (gets sorted left-to-right)
				MemPtr:= MemPtr + 2*IntSize; \reserve the memory
				Q(0):= Ptr(0);	\Nil, or pointer to larger X
				Q(1):= T;	\store the unscaled X coordinate
				Ptr(0):= Q;	\link this new entry into list
				quit;
				end
			else	Ptr:= Ptr(0);	\follow pointers in linked list
			end;
		X:= X+M;			\calculate next X coordinate
		end;
	end;	\BuildLine


begin	\DrawPoly
\Find maximum and minimum Y coordinates
N:= VT(0);				\get number of coordinates
Min:= 100000;   Max:= -100000;
I:= 2;					\index of first Y coordinate in VT
for J:= 0, N-1 do
	begin
	Y:= Y0 + VT(I);
	if Y > Max then Max:= Y;
	if Y < Min then Min:= Y;
	I:= I+2;			\next Y coordinate
	end;
Height:= Max - Min + 1;

\Set up a linked-list table. Each entry will correspond to a horizontal line
\ drawn to fill the polygon.
HeadTbl:= Reserve(Height*IntSize);
for Y:= 0, Height-1 do HeadTbl(Y):= Nil;

MemPtr:= Reserve(100000);		\set up our own heap space manager
					\ (or in-line BuildLine procedure)
I:= 1;					\index to first coordinate pair in VT
for J:= 0, N-2 do
	begin
	BuildLine(I, I+2);		\make edge of polygon between vertexes
	I:= I+2;			\ i.e. build table of X values for edge
	end;
BuildLine(I, 1);			\make sure the polygon is closed, i.e.
					\ connect last vertex to first vertex
\Draw horizontal lines to fill polygon
for Y:= Min, Max do			\from top to bottom...
	begin
	Ptr:= HeadTbl + (Y-Min)*IntSize;\get pointer to X coordinates for this Y
	loop	begin			\(loop normally exits on first pass)
		Ptr:= Ptr(0);
		if Ptr = Nil then quit;
		Gx:= Ptr(1);		\get left X coordinate for horiz line

		Ptr:= Ptr(0);
		if Ptr = Nil then quit;
		Hx:= Ptr(1);		\get right X coordinate for horiz line

		for X:= Gx+Y*ImW, Hx+Y*ImW do	\quickly draw horizontal line
			Image(X):= Color;
		end;
	end;
end;	\DrawPoly

\===============================================================================

proc	MoveTorus;	\Display and move a torus (a donut)
def	Max = 64,	\maximum limits +1 for Sectors and Sections
	Scale = 1.4,	\magnification factor (adjusted to fill screen)
	TorR1 = 100.0,	\distance from the center of the hole in the middle
			\ to the center of the eatable tube-shaped part
	TorR2 = 60.0;	\distance from the center of the eatable tube-shaped
			\ part to the frosting on top
int	Sectors,	\number of slices of the torus (tubes)
	Sections,	\number of slices of a sector (patches)
	Facets,		\number of facets
	Material,	\0=colored; 1=plaster; 2=metalic
	FPS,		\frames per second
	Sec,		\current second of time (from real-time clock)
	SecOld;		\previous second of time
real	Ambient,	\intensity of ambient light
	Light(3),	\directional light (unit vector)
	Tor(Max,Max,3),	\coordinates of vertexes
	TorN(Max,Max,3), \directions (coordinates) of normals (unit vectors)
	RTorN(Max,Max,3),\rotated normals (unit vectors)
	Ref1(3), Ref2(3), Ref3(3); \reference frame (unit vectors)
int	RTor(Max,Max,3), \rotated coordinates of vertexes
	RTorZZ(Max*Max), \decreasing Z coordinates of rotated torus
	RTorIJ(Max*Max); \corresponding I and J indexes for sorted Z coordinates

\-------------------------------------------------------------------------------

proc	Sort(AZ, AI, N);	\Sort arrays by decreasing Z coordinates
int	AZ,	\address of array of Z coordinates
	AI,	\address of array of I and J coordinates
	N;	\number of elements in array (size)
int	I, J, Gap, JG, T;
begin				\shell sort
Gap:= N>>1;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I - Gap;
		loop	begin
			JG:= J + Gap;
			if AZ(J) >= AZ(JG) then quit;
			T:= AZ(J);  AZ(J):= AZ(JG);  AZ(JG):= T;	\swap
			T:= AI(J);  AI(J):= AI(JG);  AI(JG):= T;
			J:= J - Gap;
			if J < 0 then quit;
			end;
		end;
	Gap:= Gap>>1;
	end;
end;	\Sort



proc	DepthSort;	\Sort facets by decreasing depth
\Find the Z coordinate (*4) of the approximate center of each facet. Save the
\ result in array RTorZZ. Array RTorIJ saves I & J coordinates for corresponding
\ facet. Z increases going into the screen.
\Inputs: Sectors, Sections, RTor.
\Outputs: RTorZZ, RTorIJ, Facets.
int	I, J;
begin
Facets:= 0;
for I:= 0, Sectors-1 do
    for J:= 0, Sections-1 do
	begin
	RTorZZ(Facets):= RTor(I,J,Z) + RTor(I+1,J,Z) + RTor(I+1,J+1,Z) +
			 RTor(I,J+1,Z);
	RTorIJ(Facets):= I + J<<8;
	Facets:= Facets+1;
	end;
Sort(RTorZZ, RTorIJ, Facets);
end;	\DepthSort

\-------------------------------------------------------------------------------

proc	DrawTorus;	\Draw the torus (onto Image)
\Input: Sectors, Sections, RTor, RTorN, Facets, RTorIJ, Ambient, Light, Material
int	I, J, F,	\indexes
	Color,
	PA(4*2+1);	\polygon vertexes for a (quad) facet
real	DP;
begin
PA(0):= 4;				\there are always 4 vertexes

\Draw facets starting with the farthest one (the one with the greatest Z)
DepthSort;
for F:= 0, Facets-1 do
	begin
	I:= RTorIJ(F) & $00FF;		\get indexes
	J:= RTorIJ(F) >> 8;

	\Determine if facet is visible. If its normal vector is pointing toward
	\ the viewer then it has a negative Z component and it is visible. This
	\ avoids drawing about half the polygons.
	if RTorN(I,J,Z) < 0.0 then
		begin
		case Material of
		  1:	begin		\plaster: gray range: 0..189 + 64
			DP:= DotProd(RTorN(I,J), Light);	\Lambert light
			if DP < 0.0 then DP:= 0.0;
			Color:= fix((189.0-Ambient)*DP + Ambient) + 64;
			end;
		  2:	begin		\metal: (if light source = view point)
			DP:= DotProd(RTorN(I,J), Light);
			if DP < 0.0 then DP:= 0.0;
			DP:= DP * DP;
			DP:= DP * DP;				\specular light
			Color:= fix((189.0-Ambient)*DP + Ambient) + 64;
			end
		other	Color:= rem((I+J+J)/63)+1;

		\Set up polygon (quad) vertexes for facet
		PA(1):= RTor(I,   J,   X);	PA(2):= RTor(I,   J,   Y);
		PA(3):= RTor(I+1, J,   X);	PA(4):= RTor(I+1, J,   Y);
		PA(5):= RTor(I+1, J+1, X);	PA(6):= RTor(I+1, J+1, Y);
		PA(7):= RTor(I,   J+1, X);	PA(8):= RTor(I,   J+1, Y);

		DrawPoly(ImW/2, ImH/2, PA, Color);
		end;
	end;
end;	\DrawTorus

\-------------------------------------------------------------------------------

proc	RotateLight;	\Rotate light source by moving mouse with button down
\Inputs: Light
\Outputs: Light.
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
begin
CallInt($33, $0B);		\read mouse motion counters
\Mouse X motion rotates object about Y axis
AngY:= float(-WExtend(CpuReg(2))) / 1500.0;
\Mouse Y motion rotates object about X axis
AngX:= float( WExtend(CpuReg(3))) / 1500.0;
AngZ:= 0.0;
Rotate(Light, AngX, AngY, AngZ);
end;	\RotateLight



proc	RotateTorus;	\Rotate torus by moving the mouse
\Inputs: Sectors, Sections, Scale, Tor, TorN.
\Outputs: RTor, RTorN.
int	I, J, K;		\indexes
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
begin
CallInt($33, $0B);		\read mouse motion counters
\Mouse X motion rotates object about Y axis
AngY:= float(-WExtend(CpuReg(2))) / 500.0;
\Mouse Y motion rotates object about X axis
AngX:= float( WExtend(CpuReg(3))) / 500.0;
AngZ:= 0.0;

\STEP 1:
\Rotate reference frame according to changes in mouse position.
\Independently rotating Ref1 and Ref2 can cause them to drift relative to
\ each other, but the amount of drift is insignificant because real numbers
\ are accurately calculated to 16 decimal places.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\STEP 2:
\Rotate each vertex in the object so that it aligns with the reference frame.
\Multiply each point at its initial position (Po) by the rotation matrix to get
\ the rotated position:
\					     				 
\		     	 		     Ref1(X)  Ref1(Y)  Ref1(Z) 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ref2(X)  Ref2(Y)  Ref2(Z) 
\		     	 		     Ref3(X)  Ref3(Y)  Ref3(Z) 
\					     				 
\
for I:= 0, Sectors do			\for all of the sectors+1...
    for J:= 0, Sections do		\for all of the sections+1...
	for K:= 0, 3-1 do		\(end points duplicate first points)
	    RTor(I,J,K):= fix( (Tor(I,J,X)*Ref1(K) + Tor(I,J,Y)*Ref2(K) +
				Tor(I,J,Z)*Ref3(K)) * Scale);

\Rotate each normal vector so that it aligns with the reference frame
for I:= 0, Sectors-1 do			\for all of the sectors...
    for J:= 0, Sections-1 do		\for all of the sections...
	for K:= 0, 3-1 do
	    RTorN(I,J,K):= TorN(I,J,X)*Ref1(K) + TorN(I,J,Y)*Ref2(K) +
			   TorN(I,J,Z)*Ref3(K);
end;	\RotateTorus

\-------------------------------------------------------------------------------

proc	CalcTorus;
\Find the coordinates of the vertexes of facets that approximate a torus.
\Chop the surface of the torus up into facets (polygons). To simplify later
\ calculations, the last coordinates duplicate the first coordinates. The
\ torus is initially in the X,Y plane (like a donut floating in hot grease).
\Inputs: Sectors, Sections, TorR1, TorR2.
\Outputs: Tor, TorN.
int	I, J;
real	A,		\angular increments about the hole in the center
	B,		\angular increments about the center of the tube
	SA, CA, SB, CB;	\sines and cosines of A and B
real	V1(3), V2(3);	\vectors, for finding the normals of a facet
begin
for I:= 0, Sectors do			\for all of the sectors+1...
	begin				\(end point duplicate first point)
	A:= float(I) / float(Sectors) * Pi * 2.0;
	SA:= Sin(A);   CA:= Cos(A);
	for J:= 0, Sections do		\for all of the sections+1...
		begin			\(end point duplicate first point)
		B:= float(J) / float(Sections) * Pi * 2.0;
		SB:= Sin(B);   CB:= Cos(B);
		Tor(I,J,X):= (TorR1 + TorR2*CB) * CA;
		Tor(I,J,Y):= (TorR1 + TorR2*CB) * SA;
		Tor(I,J,Z):= TorR2 * SB;
		end;
	end;

\Find the normal vectors for each facet. A normal vector is a vector that
\ is one unit long and is perpendicular to the facet. It points outward,
\ away from the inside of the donut. The inside of a facet is the side that
\ can never be seen no matter how the donut is rotated. The normal vector
\ is determined by the cross product of the vectors defined by the two
\ adjacent edges of a facet.
for I:= 0, Sectors-1 do
	for J:= 0, Sections-1 do
		begin
		VectSub(Tor(I+1,J), Tor(I,J), V1);
		VectSub(Tor(I,J+1), Tor(I,J), V2);
		CrossProd(V1, V2, TorN(I,J));
		Unitize(TorN(I,J));
		end;
end;	\CalcTorus

\-------------------------------------------------------------------------------

proc	NumOut(Line, Num);	\Display 3-digit right-justified positive number
int	Line, Num;		\ at beginning of specified Line
begin
Cursor(0, Line);
if Num < 100 then ChOut(6, ^ );
if Num < 10  then ChOut(6, ^ );
IntOut(6, Num);
end;	\NumOut

\-------------------------------------------------------------------------------

proc	Reset;			\Initialize variables to starting configuration
begin
Sectors:= 13;    Sections:= 13;
Ambient:= 39.0;
Material:= 2;
FPS:= 0;

Ref1(X):= 1.0;   Ref1(Y):= 0.0;   Ref1(Z):= 0.0;
Ref2(X):= 0.0;   Ref2(Y):= 1.0;   Ref2(Z):= 0.0;

Light(X):= -0.5; Light(Y):= -1.0; Light(Z):= -1.0;  \over left sholder of viewer
Unitize(Light);

CalcTorus;
end;	\Reset

\-------------------------------------------------------------------------------

begin	\MoveTorus
Reset;

Cursor(0, 0);			\(device 6 works for VESA graphic modes)
Text(6, "Mouse rotates torus
Button rotates light

    Sectors  [^[] [ ]

    Sections [^Y] [^X]

    Ambient  [-] [+]

    Frames/second

    Material [M]

    Reset    [Enter]

    Exit     [Esc]");

\Insert arrow character (^Z looks like an EOF, which terminates the compile)
Cursor(18,3);   ChOut(6, $1A);

loop	begin
	NumOut(3, Sectors);
	NumOut(5, Sections);
	NumOut(7, fix(Ambient/3.0));
	NumOut(11, Material);

	CallInt($1A, $0200);	\get current time in seconds
	Sec:= CpuReg(3)>>8 & $FF;
	if Sec = SecOld then	\if same as before then
		FPS:= FPS+1	\bump FPS counter
	else	begin
		SecOld:= Sec;	\save old seconds
		NumOut(9, FPS);	\display FPS counter
		FPS:= 0;	\reset FPS counter
		end;

	if MouseButton then RotateLight;
	RotateTorus;		\(make sure RTor etc. are initialized)
	EraseImage;
	DrawTorus;
	Paint(TVW-ImW, 0, ImW, ImH, Image, ImW);     \copy Image onto the screen

	if ChkKey then
		case GetKey of
		  ^M,^m:[Material:= Material+1;
			if Material > 2 then Material:= 0];
		  ^-:	[Ambient:= Ambient - 3.0;
			if Ambient<0.0 then Ambient:= 0.0];
		  ^+:	[Ambient:= Ambient + 3.0;
			if Ambient > 189.0 then Ambient:= 189.0];
		  -LtArrow:
			if Sectors > 3 then
				[Sectors:= Sectors-1;  CalcTorus];
		  -RtArrow:
			if Sectors < Max-1 then	\(end point duplicates 1st one)
				[Sectors:= Sectors+1;  CalcTorus];
		  -UpArrow:
			if Sections < Max-1 then
				[Sections:= Sections+1;  CalcTorus];
		  -DnArrow:
			if Sections > 3 then
				[Sections:= Sections-1;  CalcTorus];
		  CR:	Reset;

		  Esc:	quit
		other	ChOut(0, Bel);
	end;
end;	\MoveTorus

\===============================================================================

proc	SetColors;	\Set up color registers with 190 dithered shades of gray
int	I, R, G, B;
begin
R:= 0;	 G:= 0;   B:= 0;
port($3C8):= 64;	\start at register 64
for I:= 0, 63 do
	begin
	port($3C9):= R;  port($3C9):= G;  port($3C9):= B;
	R:= R+1;
	port($3C9):= R;  port($3C9):= G;  port($3C9):= B;
	G:= G+1;
	port($3C9):= R;  port($3C9):= G;  port($3C9):= B;
	B:= B+1;
	end;
\The last two colors are not gray because R and G are > 63
end;	\SetColors

\===============================================================================

begin	\Main
CpuReg:= GetReg;		\get address of array to access CPU registers
if not OpenMouse then [Text(0, "A mouse is required");   exit];
SetVid(VMode);
SetColors;
MoveTorus;
SetVid(3);			\restore normal text mode
end;	\Main
