\Torus6.xpl	25-Jan-2007	Boreal
\Rotatable 3D Texture-Mapped Torus
\Compile with 32-bit XPL (XPLPX, xpx.bat)
\
\A right-hand coordinate system is used:
\
\				|   Z  (into screen)
\				|  /
\				| /
\				|/
\			--------+-------- X
\			       /|
\			      / |
\			     /	|
\				|
\				Y

include	c:\cxpl\codes;		\intrinsic code definitions (including Paint)

def	VMode=$101,		\video mode: 640x480x8 graphics
	TVW=640, TVH=480,	\screen width and height, in pixels
	ImW=TVH, ImH=TVH;	\image width and height, in pixels (a square)
def	IntSize=4;		\number of bytes in an integer (2 or 4)
def	X, Y, Z;		\dimensions (indexes)

int	CpuReg,			\address of CPU register array (from GetReg)
	Gouraud;		\flag: Gouraud shading (vs. flat shading)
char	Image(ImW*ImH),		\image buffer, gets copied to the screen
	Texture(TVW*TVH),	\texture map image
	Palette(256*3),		\contains copy of color registers
	CLUT(192,256);		\color look-up table (intensity, palette number)

def	Bel=$07, CR=$0D, Esc=$1B;				\control chars
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\scan codes
def	Pi = 3.14159265358979323846;

\-------------------------------------------------------------------------------

proc	Exit(Msg);		\Make a clean exit and display a final message
char	Msg;
begin
SetVid(3);			\restore normal text mode
Text(0, Msg);
CrLf(0);
end;	\Exit



func	WExtend(N);		\Sign extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
begin
CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
SoftInt(Int);
return WExtend(CpuReg(0));	\return value in AX register (as 32 bits)
end;	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
begin				\ no Ctrl+C, and no cursor
SC:= CallInt($16, $0000);	\BIOS function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
end;	\GetKey



func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\Pointer is set to center of screen but hidden
CallInt($21, $3533);		\Make sure mouse vector $33 points to something
if WExtend(CpuReg(1))=0 & WExtend(CpuReg(11))=0 then return false;
CallInt($33, $21);		\reset mouse for Windows
CallInt($33, $0B);		\read motion counters to reset them for Windows
return CallInt($33, $0000);	\h/w reset mouse and return status
end;	\OpenMouse

\-------------------------------------------------------------------------------

proc	VectSub(V1, V2, V3);	\Subtract two 3D vectors
real	V1, V2, V3;		\V3:= V1 - V2
begin
V3(X):= V1(X) - V2(X);
V3(Y):= V1(Y) - V2(Y);
V3(Z):= V1(Z) - V2(Z);
end;	\VectSub



proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
begin
V3(X):= V1(Y)*V2(Z) - V1(Z)*V2(Y);
V3(Y):= V1(Z)*V2(X) - V1(X)*V2(Z);
V3(Z):= V1(X)*V2(Y) - V1(Y)*V2(X);
end;	\CrossProd



func real DotProd(V1, V2);	\Return the dot product of two 3D vectors
real	V1, V2;			\V1  V2
return V1(X)*V2(X) + V1(Y)*V2(Y) + V1(Z)*V2(Z);



proc	Unitize(V);		\Convert a 3D vector to its unit vector
real	V;			\vector to convert
real	M;			\magnitude of vector
begin
M:= Sqrt( V(X)*V(X) + V(Y)*V(Y) + V(Z)*V(Z) );
V(X):= V(X)/M;
V(Y):= V(Y)/M;
V(Z):= V(Z)/M;
end;	\Unitize



proc	Rotate(V, R, P, W); \3D rotate vector V
real	V,		\3D vector is 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	SW, SP, SR, CW, CP, CR, T;
begin
SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= V(Y)*CR - V(Z)*SR;		\		       	 
V(Z):= V(Z)*CR + V(Y)*SR;	\		      1   0   0 
V(Y):= T;			\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= V(X)*CP + V(Z)*SP;		\		       	 
V(Z):= V(Z)*CP - V(X)*SP;	\		      CP  0  -SP
V(X):= T;			\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= V(X)*CW - V(Y)*SW;		\		       	  
V(Y):= V(Y)*CW + V(X)*SW;	\		       CW  SW	0 
V(X):= T;			\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
end;	\Rotate					       	  



proc	EraseImage;		\Erase the Image array
int	Im, I;
begin
Im:= Image;			\erase 4 bytes at a time, for speed
for I:= 0, (ImW*ImH)>>2-1 do Im(I):= 0;
end;

\-------------------------------------------------------------------------------

proc	DrawPoly(X0, Y0, VT, CT, TT);	\Draw texture-mapped polygon onto Image
\Inputs: Gouraud, Texture, CLUT.
\This works for convex as well as concave polygons, and even for polygons
\ containing holes. Initially a table is set up that defines the coordinates
\ along the edges of the polygon. Each location in the table corresponds to a
\ Y coordinate, and (in the simple, usual case) it contains the corresponding
\ left and right X coordinates of the polygon's edges. After the table is set
\ up, horizontal lines are drawn to fill the polygon. Ref: "Computer Graphics:
\ Principals and Practice, Second Edition" by John Foley, et al.

int	X0, Y0,		\position to draw polygon (pixels)
	VT,		\vertex table: 1-dimensional array of X,Y coordinates
			\ first value gives the number of vertexes (X,Y pairs)
	CT,		\color table: color intensities at vertexes
	TT;		\texture table: 1-dimensional array of U,V coordinates
int	N, HeadTbl, Ptr, MemPtr, \used to build linked-list table
	Gx, Gy, Hx, Hy,	\coordinates of end points of drawn horizontal scan line
	Gc, Hc,		\color intensity at left and right side of scan line
	CI,		\color increment for (scaled) color intensity
	Int,		\intensity of light, if flat shading is selected
	I, J,		\indexes
	X, Y,		\coordinates used to draw horizontal scan lines
	DX,		\delta X
	U, V,		\coordinates in texture map
	Mu, Mv,		\increments in texture map
	Min, Max,	\minimum and maximum Y coordinates of polygon
	Height;		\height of polygon, in pixels
def	Nil= $80000000,	\linked list null pointer
	SS = 16; 	\scale shift factor


	proc	BuildLine(I0, I1);
	\Build a table at MemPtr that contains X coordinates and light
	\ intensities along the polygon's edge. This makes a linked list
	\ for each horizontal line that will be drawn to fill the polygon.
	int	I0, I1;	\indexes in VT specifying vertexes to build line between
	int	X, Y, DY, M, T, Q, C;
	begin
	Gx:= VT(I0) + X0;   Gy:= VT(I0+1) + Y0;	\get coordinates of end points
	Hx:= VT(I1) + X0;   Hy:= VT(I1+1) + Y0;	\ of the polygon's edge
	Gc:= CT(I0>>1);     Hc:= CT(I1>>1);

	if Gy > Hy then
		begin				\swap end points, so that Hy is
		T:= Gy;   Gy:= Hy;   Hy:= T;	\ lower than (or equal to) Gy
		T:= Gx;   Gx:= Hx;   Hx:= T;
		T:= Gc;   Gc:= Hc;   Hc:= T;
		T:= I0;   I0:= I1;   I1:= T;
		end;

	DY:= Hy - Gy;				\delta Y
	if DY = 0 then return;			\horizontal edges can be ignored
	X:= Gx<<SS;				\scaled starting X coordinate
	M:= (Hx-Gx)<<SS / DY;			\scaled slope (rotated 90 deg)

	C:= Gc<<SS;
	CI:= (Hc-Gc)<<SS / DY;			\interpolate color along Y axis

	U:= TT(I0)<<SS;				\get texture coords at start of
	V:= TT(I0+1)<<SS;			\ edge and scale them up
	Mu:= (TT(I1)<<SS - U) / DY;		\increments for stepping through
	Mv:= (TT(I1+1)<<SS - V) / DY;		\ the texture image (slopes)

	for Y:= Gy-Min, Hy-Min-1 do		\follow the edge downward...
		begin				\(bottom-most point isn't drawn)
		T:= X/(1<<SS);			\unscale signed X coordinate
		Ptr:= HeadTbl + Y*IntSize;	\pointer from list header table
		loop	begin
			if Ptr(0)=Nil ! Ptr(0,1)>T then
				begin		\insert item at front of list
				Q:= MemPtr;	\ (gets sorted left-to-right)
				MemPtr:= MemPtr + 5*IntSize; \reserve the memory
				Q(0):= Ptr(0);	\Nil, or pointer to larger X
				Q(1):= T;	\store the unscaled X coordinate
				Q(2):= C;	\store scaled color intensity
				Q(3):= U;	\store scaled texture coords
				Q(4):= V;
				Ptr(0):= Q;	\link this new entry into list
				quit;
				end
			else	Ptr:= Ptr(0);	\follow pointers in linked list
			end;
		X:= X+M;			\calculate next X coordinate
		C:= C+CI;			\next color intensity
		U:= U+Mu;			\next coordinates in texture map
		V:= V+Mv;
		end;
	end;	\BuildLine


begin	\DrawPoly
Int:= CT(0);
\Find maximum and minimum Y coordinates
N:= VT(0);				\get number of coordinates
Min:= 100000;   Max:= -100000;
I:= 2;					\index of first Y coordinate in VT
for J:= 0, N-1 do
	begin
	Y:= Y0 + VT(I);
	if Y > Max then Max:= Y;
	if Y < Min then Min:= Y;
	I:= I+2;			\next Y coordinate
	end;
if Min>=ImH ! Max<0 then return;	\clip whole polygon
Height:= Max - Min + 1;

\Set up a linked-list table. Each entry will correspond to a horizontal line
\ drawn to fill the polygon.
HeadTbl:= Reserve(Height*IntSize);
for Y:= 0, Height-1 do HeadTbl(Y):= Nil;

MemPtr:= Reserve(200000);		\set up our own heap space manager
					\ (or in-line BuildLine procedure)
I:= 1;					\index to first coordinate pair in VT
for J:= 0, N-2 do
	begin
	BuildLine(I, I+2);		\make edge of polygon between vertexes
	I:= I+2;			\ i.e. build table of X values for edge
	end;
BuildLine(I, 1);			\make sure the polygon is closed, i.e.
					\ connect last vertex to first vertex
\Draw horizontal lines to fill polygon
if Max >= ImH then Max:= ImH-1;		\clip partial polygon
for Y:= if Min<0 then 0 else Min, Max do \from top to bottom...
	begin
	Ptr:= HeadTbl + (Y-Min)*IntSize;\get pointer to X coordinates for this Y
	loop	begin			\(loop normally exits on first pass)
		Ptr:= Ptr(0);
		if Ptr = Nil then quit;
		Gx:= Ptr(1);		\get left X coordinate for horiz line
		Gc:= Ptr(2);		\get starting color intensity
		U:= Ptr(3);		\get corresponding texture coordinates
		V:= Ptr(4);

		Ptr:= Ptr(0);
		if Ptr = Nil then quit;
		Hx:= Ptr(1);		\get right X coordinate for horiz line
		Hc:= Ptr(2);		\get ending color intensity

		if Gx>=ImW ! Hx<0 then quit;	\clip entire scan line

		DX:= Hx - Gx;
		if DX \#0\ then
			begin
			if Gouraud then
			    CI:= (Hc-Gc) / (Hx-Gx);   \interpolate color along X
			Mu:= (Ptr(3) - U) / DX;	\increments for stepping thru
			Mv:= (Ptr(4) - V) / DX;	\ texture image
			end;
		if Gx < 0 then
			begin		\clip
			if Gouraud then
			    Gc:= Gc - Gx*CI; \adjust color for clipped scan line
			U:= U - Gx*Mu;	\compute new starting coordinates in
			V:= V - Gx*Mv;	\ texture image
			Gx:= 0;		\clip horiz line to left edge of Image
			end;
		if Hx >= ImW then Hx:= ImW-1;	\clip to right edge of Image

		if Gouraud then
		    for X:= Gx+Y*ImW, Hx+Y*ImW do  \quickly draw horizontal line
			begin
			Image(X):= CLUT(Gc>>SS, Texture(U>>SS + (V>>SS)*TVW));
			Gc:= Gc + CI;	\adjust color intensity
			U:= U + Mu;	\step coordinates thru texture image
			V:= V + Mv;
			end
		else
		    for X:= Gx+Y*ImW, Hx+Y*ImW do  \quickly draw horizontal line
			begin
			Image(X):= CLUT(Int, Texture(U>>SS + (V>>SS)*TVW));
			U:= U + Mu;	\step coordinates thru texture image
			V:= V + Mv;
			end;
		end;
	end;
end;	\DrawPoly

\===============================================================================

proc	MoveTorus;	\Display and move a torus (a donut)
def	Max = 64,	\maximum limits +1 for Sectors and Sections
	TorR1 = 100.0,	\distance from the center of the hole in the middle
			\ to the center of the eatable tube-shaped part
	TorR2 = 60.0;	\distance from the center of the eatable tube-shaped
			\ part to the frosting on top
int	Sectors,	\number of slices of the torus (tubes)
	Sections,	\number of slices of a sector (patches)
	Facets,		\number of facets
	FPS,		\frames per second
	Sec,		\current second of time (from real-time clock)
	SecOld;		\previous second of time
real	Ambient,	\intensity of ambient light
	Light(3),	\directional light (unit vector)
	Scale,		\magnification factor
	Tor(Max,Max,3),	\coordinates of vertexes
	TorN(Max,Max,3), \directions (coordinates) of normals (unit vectors)
	RTorN(Max,Max,3),\rotated normals (unit vectors)
        TorVN(Max,Max,3), \vertex normals (unit vectors)
        RTorVN(Max,Max,3),\rotated vertex normals (unit vectors)
	Ref1(3), Ref2(3), Ref3(3); \reference frame (unit vectors)
int	Offset(3),	 \offset position to add to vertexes (X,Y,Z)
	RTor(Max,Max,3), \rotated coordinates of vertexes
	RTorZZ(Max*Max), \decreasing Z coordinates of rotated torus
	RTorIJ(Max*Max); \corresponding I and J indexes for sorted Z coordinates

\-------------------------------------------------------------------------------

proc	Sort(AZ, AI, N);	\Sort arrays by decreasing Z coordinates
int	AZ,	\address of array of Z coordinates
	AI,	\address of array of I and J coordinates
	N;	\number of elements in array (size)
int	I, J, Gap, JG, T;
begin				\shell sort
Gap:= N>>1;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I - Gap;
		loop	begin
			JG:= J + Gap;
			if AZ(J) >= AZ(JG) then quit;
			T:= AZ(J);  AZ(J):= AZ(JG);  AZ(JG):= T;	\swap
			T:= AI(J);  AI(J):= AI(JG);  AI(JG):= T;
			J:= J - Gap;
			if J < 0 then quit;
			end;
		end;
	Gap:= Gap>>1;
	end;
end;	\Sort



proc	DepthSort;	\Sort facets by decreasing depth
\Find the Z coordinate (*4) of the approximate center of each facet. Save the
\ result in array RTorZZ. Array RTorIJ saves I & J coordinates for corresponding
\ facet. Z increases going into the screen.
\Inputs: Sectors, Sections, RTor.
\Outputs: RTorZZ, RTorIJ, Facets.
int	I, J;
begin
Facets:= 0;
for I:= 0, Sectors-1 do
    for J:= 0, Sections-1 do
	begin
	RTorZZ(Facets):= RTor(I,J,Z) + RTor(I+1,J,Z) + RTor(I+1,J+1,Z) +
			 RTor(I,J+1,Z);
	RTorIJ(Facets):= I + J<<8;
	Facets:= Facets+1;
	end;
Sort(RTorZZ, RTorIJ, Facets);
end;	\DepthSort

\-------------------------------------------------------------------------------

proc	DrawTorus;	\Draw the torus (onto Image)
\Input: Sectors, Sections, RTor, RTorN, Facets, RTorIJ, Ambient, Light.
int	I, J, K, F,	\indexes
	TI, TJ,		\tables to select vertex coordinates in sequence
	VT(4*2+1),	\polygon vertex table for a (quad) facet
	CT(4),		\color table: color intensities at vertexes
	TT(4*2+1);	\texture table
real	DP;		\dot product result
begin
TI:= [0, 1, 1, 0];
TJ:= [0, 0, 1, 1];
VT(0):= 4;				\there are always 4 vertexes

\Draw facets starting with the farthest one (the one with the greatest Z)
DepthSort;
for F:= 0, Facets-1 do
	begin
	I:= RTorIJ(F) & $00FF;		\get indexes
	J:= RTorIJ(F) >> 8;

	\Determine if facet is visible. If its normal vector is pointing toward
	\ the viewer then it has a negative Z component and it is visible. This
	\ avoids drawing about half the polygons.
	if RTorN(I,J,Z) < 0.0 then
		begin
		if Gouraud then
		    for K:= 0, 3 do
			begin		\brightness range: 0..189
			DP:= DotProd(RTorVN(I+TI(K), J+TJ(K)), Light);
			if DP < 0.0 then DP:= 0.0;		\Lambert light
			CT(K):= Fix((189.0-Ambient)*DP + Ambient);
			end
		else	begin		\flat shading
			DP:= DotProd(RTorN(I,J), Light);
			if DP < 0.0 then DP:= 0.0;
			CT(0):= Fix((189.0-Ambient)*DP + Ambient);
			end;

		\Set up polygon (quad) vertexes for facet
		VT(1):= RTor(I,   J,   X);	VT(2):= RTor(I,   J,   Y);
		VT(3):= RTor(I+1, J,   X);	VT(4):= RTor(I+1, J,   Y);
		VT(5):= RTor(I+1, J+1, X);	VT(6):= RTor(I+1, J+1, Y);
		VT(7):= RTor(I,   J+1, X);	VT(8):= RTor(I,   J+1, Y);

		\Set up corresponding coordinates in texture image
		for K:= 0, 3 do
			begin
			TT(K*2+1):= (I+TI(K))*627/Sectors;
			TT(K*2+2):= (J+TJ(K))*377/Sections;
			end;

                DrawPoly(ImW/2, ImH/2, VT, CT, TT);
		end;
	end;
end;	\DrawTorus

\-------------------------------------------------------------------------------

proc	RotateLight;	\Rotate light source by moving mouse with button down
\Inputs: Light
\Outputs: Light.
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
begin
CallInt($33, $0B);		\read mouse motion counters
\Mouse X motion rotates object about Y axis
AngY:= Float(-WExtend(CpuReg(2))) / 1500.0;
\Mouse Y motion rotates object about X axis
AngX:= Float( WExtend(CpuReg(3))) / 1500.0;
AngZ:= 0.0;
Rotate(Light, AngX, AngY, AngZ);
end;	\RotateLight



proc	RotateTorus;	\Rotate torus by moving the mouse
\Inputs: Sectors, Sections, Scale, Tor, TorN.
\Outputs: RTor, RTorN.
int	I, J, K;		\indexes
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
begin
AngY:= 0.0;
AngX:= 0.0;
AngZ:= 0.0;

CallInt($33, $0003);
case CpuReg(1) & $03 of
  $02:	begin	\right button
	CallInt($33, $0B);		\read mouse motion counters
	Offset(X):= Offset(X) + WExtend(CpuReg(2));
	Offset(Y):= Offset(Y) + WExtend(CpuReg(3));
	end;
  $03:	begin	\both buttons
	CallInt($33, $0B);		\read mouse motion counters
	Scale:= Scale + Float(-WExtend(CpuReg(3)))/1000.0;
	end
other	begin	\no buttons
	CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Y axis
	AngY:= Float(-WExtend(CpuReg(2))) / 500.0;
	\Mouse Y motion rotates object about X axis
	AngX:= Float( WExtend(CpuReg(3))) / 500.0;
	AngZ:= 0.0;
	end;

\STEP 1:
\Rotate reference frame according to changes in mouse position.
\Independently rotating Ref1 and Ref2 can cause them to drift relative to
\ each other, but the amount of drift is insignificant because real numbers
\ are accurately calculated to 16 decimal places.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2


\STEP 2:
\Rotate each vertex in the object so that it aligns with the reference frame.
\Multiply each point at its initial position (Po) by the rotation matrix to get
\ the rotated position:
\					     				 
\		     	 		     Ref1(X)  Ref1(Y)  Ref1(Z) 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ref2(X)  Ref2(Y)  Ref2(Z) 
\		     	 		     Ref3(X)  Ref3(Y)  Ref3(Z) 
\					     				 
\
for I:= 0, Sectors do			\for all of the sectors+1...
    for J:= 0, Sections do		\for all of the sections+1...
	for K:= 0, 3-1 do		\(end points duplicate first points)
	    RTor(I,J,K):= Fix( (Tor(I,J,X)*Ref1(K) + Tor(I,J,Y)*Ref2(K) +
				Tor(I,J,Z)*Ref3(K)) * Scale) + Offset(K);

\Rotate each normal vector so that it aligns with the reference frame
for I:= 0, Sectors-1 do			\for all of the sectors...
    for J:= 0, Sections-1 do		\for all of the sections...
	for K:= 0, 3-1 do
	    RTorN(I,J,K):= TorN(I,J,X)*Ref1(K) + TorN(I,J,Y)*Ref2(K) +
			   TorN(I,J,Z)*Ref3(K);

\Rotate vertex normals
if Gouraud then
  for I:= 0, Sectors do			\for all of the sectors+1...
    for J:= 0, Sections do		\for all of the sections+1...
	for K:= 0, 3-1 do		\(end points duplicate first points)
	    RTorVN(I,J,K):= TorVN(I,J,X)*Ref1(K) + TorVN(I,J,Y)*Ref2(K) +
			    TorVN(I,J,Z)*Ref3(K);
end;	\RotateTorus

\-------------------------------------------------------------------------------

proc	CalcTorus;
\Find the coordinates of the vertexes of facets that approximate a torus.
\Chop the surface of the torus up into facets (polygons). To simplify later
\ calculations, the last coordinates duplicate the first coordinates. The
\ torus is initially in the X,Y plane (like a donut floating in hot grease).
\Inputs: Sectors, Sections, TorR1, TorR2.
\Outputs: Tor, TorN, TorVN.
int     I, J, K, I1, J1, I2, J2;
real	A,		\angular increments about the hole in the center
	B,		\angular increments about the center of the tube
	SA, CA, SB, CB;	\sines and cosines of A and B
real	V1(3), V2(3);	\vectors, for finding the normals of a facet
begin
for I:= 0, Sectors do			\for all of the sectors+1...
	begin				\(end point duplicate first point)
	A:= Float(I) / Float(Sectors) * Pi * 2.0;
	SA:= Sin(A);   CA:= Cos(A);
	for J:= 0, Sections do		\for all of the sections+1...
		begin			\(end point duplicate first point)
		B:= Float(J) / Float(Sections) * Pi * 2.0;
		SB:= Sin(B);   CB:= Cos(B);
		Tor(I,J,X):= (TorR1 + TorR2*CB) * CA;
		Tor(I,J,Y):= (TorR1 + TorR2*CB) * SA;
		Tor(I,J,Z):= TorR2 * SB;
		end;
	end;

\Find the normal vectors for each facet. A normal vector is a vector that
\ is one unit long and is perpendicular to the facet. It points outward,
\ away from the inside of the donut. The inside of a facet is the side that
\ can never be seen no matter how the donut is rotated. The normal vector
\ is determined by the cross product of the vectors defined by the two
\ adjacent edges of a facet.
for I:= 0, Sectors-1 do
	for J:= 0, Sections-1 do
		begin
		VectSub(Tor(I+1,J), Tor(I,J), V1);
		VectSub(Tor(I,J+1), Tor(I,J), V2);
		CrossProd(V1, V2, TorN(I,J));
		Unitize(TorN(I,J));
		end;

\Find the vertex normals by averaging the normals of the adjacent facets
if Gouraud then
    for I:= 0, Sectors do
	begin
	I1:= I-1;
	if I1 < 0 then I1:= Sectors-1;
	I2:= I;
	if I2 = Sectors then I2:= 0;
	for J:= 0, Sections do
		begin
		J1:= J-1;
		if J1 < 0 then J1:= Sections-1;
		J2:= J;
		if J2 = Sections then J2:= 0;
		for K:= 0, 3-1 do
			TorVN(I,J,K):= TorN(I2,J2,K) + TorN(I1,J2,K) +
				       TorN(I1,J1,K) + TorN(I2,J1,K);
		Unitize(TorVN(I,J));
		end;
	end;
end;	\CalcTorus

\-------------------------------------------------------------------------------

proc	NumOut(Line, Num);	\Display 3-digit right-justified positive number
int	Line, Num;		\ at beginning of specified Line
begin
Cursor(0, Line);
if Num < 100 then ChOut(6, ^ );
if Num < 10  then ChOut(6, ^ );
IntOut(6, Num);
end;	\NumOut

\-------------------------------------------------------------------------------

proc	Reset;			\Initialize variables to starting configuration
begin
Sectors:= 13;    Sections:= 13;
Scale:= 1.4;			\magnification factor (adjusted to fill screen)
Offset(X):= 0;   Offset(Y):= 0;   Offset(Z):= 0;
Ambient:= 39.0*3.0;
Gouraud:= true;
FPS:= 0;

Ref1(X):= 1.0;   Ref1(Y):= 0.0;   Ref1(Z):= 0.0;
Ref2(X):= 0.0;   Ref2(Y):= 1.0;   Ref2(Z):= 0.0;

Light(X):= -0.5; Light(Y):= -1.0; Light(Z):= -1.0;  \over left sholder of viewer
Unitize(Light);

CalcTorus;
end;	\Reset

\-------------------------------------------------------------------------------

begin	\MoveTorus
Reset;

Cursor(0, 0);			\(device 6 works for VESA graphic modes)
Text(6, "Mouse rotates torus
Left button, light
Right button pans
Both buttons zoom

    Sectors  [^[] [ ]

    Sections [^Y] [^X]

    Ambient  [-] [+]

    Frames/second

    Gouraud  [G]

    Reset    [Enter]

    Exit     [Esc]");

\Insert arrow character (^Z looks like an EOF, which terminates the compile)
Cursor(18,5);   ChOut(6, $1A);

loop	begin
	NumOut(5, Sectors);
	NumOut(7, Sections);
	NumOut(9, Fix(Ambient/3.0));
	NumOut(13, if Gouraud then 1 else 0);

	CallInt($1A, $0200);	\get current time in seconds
	Sec:= CpuReg(3)>>8 & $FF;
	if Sec = SecOld then	\if same as before then
		FPS:= FPS+1	\bump FPS counter
	else	begin
		SecOld:= Sec;	\save old seconds
		NumOut(11, FPS);\display FPS counter
		FPS:= 0;	\reset FPS counter
		end;

	CallInt($33, $0003);
	if (CpuReg(1)&$03) = $01 \left\ then RotateLight;
	RotateTorus;		\(make sure RTor etc. are initialized)
	EraseImage;
	DrawTorus;
	Paint(TVW-ImW, 0, ImW, ImH, Image, ImW);     \copy Image onto the screen

	if ChkKey then
		case GetKey of
		  ^G,^g:[Gouraud:= ~Gouraud;  CalcTorus];
		  ^-:	[Ambient:= Ambient - 3.0;
			if Ambient<0.0 then Ambient:= 0.0];
		  ^+:	[Ambient:= Ambient + 3.0;	\(189/3 = 63.0)
			if Ambient > 189.0 then Ambient:= 189.0];
		  -LtArrow:
			if Sectors > 3 then
				[Sectors:= Sectors-1;  CalcTorus];
		  -RtArrow:
			if Sectors < Max-1 then	\(end point duplicates 1st one)
				[Sectors:= Sectors+1;  CalcTorus];
		  -UpArrow:
			if Sections < Max-1 then
				[Sections:= Sections+1;  CalcTorus];
		  -DnArrow:
			if Sections > 3 then
				[Sections:= Sections-1;  CalcTorus];
		  CR:	Reset;

		  Esc:	quit
		other	ChOut(0, Bel);
	end;
end;	\MoveTorus

\===============================================================================

proc	MakeCLUT;	\Make Color Look-Up Table
\Inputs: Palette.
\Outputs: CLUT
int	CReg,		\color register number [0..255]
	R, G, B,	\red, green, blue varying intensities [0..63]
	R0, G0, B0,	\original red, green, blue intensities
	DR, DG, DB,	\delta red, green, blue
	D,		\total delta
	Min,		\minimum delta
	I, J, J3, SJ;	\indexes


	proc	Search;
	\For each color intensity find the closest available color
	begin
	Min:= $7FFFFFFF;
	for J:= 0, 255 do
		begin
		J3:= J+J+J;
		DR:= Palette(J3+0) - R;
		DG:= Palette(J3+1) - G;
		DB:= Palette(J3+2) - B;

		D:= DR*DR + DG*DG + DB*DB;
		if D < Min then
			begin
			SJ:= J;		\save index of closest color
			Min:= D;	\record error for best match
			end;
		end;
	end;	\Search


begin	\MakeCLUT
for CReg:= 0, 255 do				\for all 256 colors...
	begin
	R0:= Palette(CReg*3+0);
	G0:= Palette(CReg*3+1);
	B0:= Palette(CReg*3+2);

	G:= 0;   B:= 0;
	for I:= 0, 63 do			\for 64*3 intensity levels...
		begin
		R:= (R0*I) >> 6;		\ = R0 * I/64
		Search;
		\Store closest color register number into CLUT
		CLUT(I*3+0, CReg):= SJ;

		G:= (G0*I) >> 6;
		Search;
		CLUT(I*3+1, CReg):= SJ;

		B:= (B0*I) >> 6;
		Search;
		CLUT(I*3+2, CReg):= SJ;
		end;
	end;
end;	\MakeCLUT



proc	LoadBmp;	\Load the texture.bmp file and set up color registers
\Outputs: Texture, Palette.
int	Hand,		\file handle
	I,		\index
	T,		\temporary scratch
	X, Y,		\coordinates
	W, H,		\width, height
	R, G, B;	\red, green, blue
begin
Trap(false);			\don't trap "file not found" error; we'll do it
Hand:= FOpen("TEXTURE.BMP", 0);	\open file for input
if GetErr \#0\ then Exit("TEXTURE.BMP not found");
Trap(true);
FSet(Hand, ^I);			\set device 3 to handle
OpenI(3);

for Y:= 0, 17 do X:= ChIn(3);	\skip unused header info
W:= ChIn(3) + ChIn(3)<<8;	\0..32764 (ample range)
for Y:= 0, 2-1 do X:= ChIn(3);	\skip
H:= ChIn(3) + ChIn(3)<<8;	\0..32767 (ample range)
for Y:= 24, 53 do X:= ChIn(3);	\skip

port($3C8):= 0;			\set color registers
for I:= 0, 255 do
	begin
	B:= ChIn(3)>>2;
	G:= ChIn(3)>>2;
	R:= ChIn(3)>>2;
	T:= ChIn(3);
	port($3C9):= R;
	port($3C9):= G;
	port($3C9):= B;
	Palette(3*I+0):= R;
	Palette(3*I+1):= G;
	Palette(3*I+2):= B;
	end;

for Y:= -(H-1), 0 do	\.bmp files are upside down
    for X:= 0, W-1 do
	Texture(X-Y*W):= ChIn(3);
FClose(Hand);
end;	\LoadBmp

\===============================================================================

begin	\Main
CpuReg:= GetReg;		\get address of array to access CPU registers
if not OpenMouse then [Text(0, "A mouse is required");   exit];
SetVid(VMode);
LoadBmp;
MakeCLUT;
MoveTorus;
SetVid(3);			\restore normal text mode
end;	\Main
