/*
	Public Domain Code
	Coded by The IceBreaker
*/

#ifndef __model3ds_h__
#define __model3ds_h__

#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <stdio.h>

/* it is more convenient to define our own boolean :D */
typedef enum {false,true} nboolean;

/* magic numbers from the 3DS specification */
#define MAIN_CHUNK               0x4d4d
#define VERSION                  0x0002
#define EDITOR_CHUNK             0x3d3d
#define OBJECT_BLOCK             0x4000
#define TRIANGULAR_MESH          0x4100
#define VERTICES_LIST            0x4110
#define FACES_DESCRIPTION        0x4120
#define FACES_MATERIAL_LIST		 0x4130
#define MAPPING_COORDINATES_LIST 0x4140
/* 
	As you probably know the smallest element
	of the 3D space is the vertex.
	
	A vertex usually is defined by 3 coordinates,
	but this time we include the texture coordinates
	too, commonly named U and V.
*/
typedef struct t_Vertex
{
	float x,y,z;
	float u,v;
} Vertex;
/*
	A face (polygon,triangle) contains in our case indices
	to 3 vertices. We are storing indices because this way
	multiple faces can share the same vertices. A 3DS file
	always stores faces AS triangles (and so do most of the
	3D file formats) so the number of indices is always three.
	
	If you will want to implement your own file format then
	it is useful to make the number of indices variable.
*/
typedef struct t_Face 
{
	unsigned short a,b,c;	
} Face;
/*
	A mesh is a collection of Vertices and Faces, but it also
	holds a name. These are stored in simple dynamic arrays.
	
	Storing the number of vertices respectivly faces, it is
	not that important because this number could be calculated
	from the allocated array, but it is more convenient this way.
	
	To calculate the length of a dynamic array you do this:
	int length = (sizeof(myarray[0])/sizeof(myarray));
	
	In *human* language, you divide the size of the first element
	with the size of the whole array.
*/
typedef struct t_Mesh
{
	char MeshName[64];
	unsigned short NumVertices;
	unsigned short  NumFaces;
	
	Vertex *Vertices;
	Face   *Faces;
} Mesh;
/*
	As I said before a 3DS file can contain multiple meshes, so the
	Model structure holds a dynamic array of meshes, a model name,
	and two vectors to positionate and rotate the model.
	
	The last two are not required for the loading part, but for
	rendering.
*/
typedef struct t_Model
{
	char ModelName[64];
	
	/* Added for rendering */
	float position[3];
	float rotation[3];
	
	unsigned short NumMeshes;
	Mesh *Meshes;
} Model;
/*
	To see the implementation of these functions, proceed to
	the model3ds.c implementation file.
*/
nboolean 	model_Load	(Model *model, const char *FileName);
void	 	model_Free	(Model *model);
void		model_Render(Model *model);
void		model_Rotate(Model *model, float x, float y, float z);

#endif /* !__model3ds_h__ */
